

module RefacAddRmParam(addOneParameter,rmOneParameter) where

import PosSyntax
import TypedIds
import UniqueNames
import PNT 
import TiPNT 

import Maybe
import List  
import RefacUtils
import Char


-----------------------------------------------------------------------------------------------------
{- An argument can be added to the definition of a function or constant. Adding an argument to a constant 
   definition will change the constant definition to a function definition.  The new parameter is always
   added as the first parameter of the function. A default parameter will be added as the first argument
   to each of the function's call site. Suppose a new  parameter named 'p' is added to function 'f',
   then default parameter will be defined automatically as p_f_i=undefined, where 'i' is an integer.
   To ensure that the default parameter name does not cause name clash in the client modules, we take the
   visble names both in the current module and in the client modules into account when creating the the
   name. 

-}
----------------------------------------------------------------------------------------------------- 
addOneParameter args
 = let fileName = args!!0
       --fileName'= moduleName fileName
       paramName= args!!1          
       row      = read (args!!2)::Int
       col      = read (args!!3)::Int
   in if isVarId paramName
       then do modName <- RefacUtils.moduleName fileName  
               (inscps, exps, mod, tokList)<-parseSourceFile fileName  
               let pnt=locToPNT fileName  row col mod
                   pn =pNTtoPN pnt 
               --make sure this name is defined in this module
               if pn /= defaultPN && isFunOrPatName pn mod
                 then if isExported pnt exps
                       then do clients <- clientModsAndFiles modName
                               info    <- mapM parseSourceFile $ map snd  clients         
                               defaultArg <-mkTopLevelDefaultArgName pn paramName fileName modName 
                                                    (inScopeNames (map myfst info)) (declsIn mod)
                               (mod',((tokList',m),_))<-doAddingParam fileName  modName pn paramName (Just defaultArg) True  mod tokList   
                               refactoredClients<-mapM (addArgInClientMod pnt defaultArg modName) $ zip info (map snd clients)
                               writeRefactoredFiles $ ((fileName,m),(tokList',mod')):refactoredClients
                       else do (mod',((tokList',m),_))<-doAddingParam fileName  modName pn paramName Nothing  False mod tokList 
                               writeRefactoredFiles [((fileName,m),(tokList',mod'))]                               
                 else error "Invalid cursor position or the identifier is not a function/pattern name defined in this module!\n"
       else error "Invalid parameter name!\n"

doAddingParam fileName modName pn newParam defaultArg isExported mod tokList  
    =  runStateT (applyTP (once_tdTP (failTP `adhocTP` inMod
                                             `adhocTP` inMatch
                                             `adhocTP` inPat
                                             `adhocTP` inLet
                                             `adhocTP` inAlt
                                             `adhocTP` inLetStmt)
                           `choiceTP` failure) mod) ((tokList,unmodified),(fileName,initValue-stepValue))
      where
             --1.pn is declared in top level
             inMod (mod@(HsModule loc name exps imps ds):: HsModuleP)  
               | findDecl pn ds False 
               = do mod'<-doAdding mod  ds   
                    if isExported && isExplictlyExported pn mod
                      then addVarItemInExport mod' pn  (fromJust defaultArg)
                      else return mod' 
                            
             inMod _ = mzero 
           
             --2. pn is declared locally in the where clause of a match.
             inMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
               | findDecl pn ds False = doAdding match  ds                                                
             inMatch _ = mzero

             --3. pn is declared locally in the where clause of a pattern binding.
             inPat (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
               | findDecl pn ds False = doAdding pat  ds                  
             inPat _ = mzero
  
             --4: pn is declared locally in a  Let expression
             inLet (letExp@(Exp (HsLet ds e))::HsExpP)
               | findDecl pn  ds False = doAdding letExp  ds 
             inLet _ = mzero
                
             --5. pn is declared locally in a  case alternative.
             inAlt (alt@(HsAlt loc p rhs ds)::HsAltP)
               | findDecl pn ds False = doAdding  alt  ds 
             inAlt _ = mzero
      
             --6.pn is declared locally in a let statement.
             inLetStmt (letStmt@(HsLetStmt ds stmts):: HsStmtP)
               | findDecl pn ds False = doAdding letStmt ds 
             inLetStmt _ = mzero
            
             failure = idTP `adhocTP` mod
               where mod (m::HsModuleP) = error "Refactoring failed"
 
             doAdding parent ds  
               = do ds' <- addFormalParam pn newParam ds
                    defaultParamPName <-if isNothing defaultArg
                                           then mkLocalDefaultArgName pn newParam modName parent                
                                           else return (fromJust defaultArg)
                    parent' <- addDefaultActualArg False pn defaultParamPName (replaceDecls parent ds')
                    ds'' <- addArgToSig pn =<< addDefaultActualArgDecl defaultParamPName  parent' pn isExported
                    return (replaceDecls parent' ds'' ) 
         
--all the new formal parameter to the declaration defining pn
addFormalParam pn newParam = applyTP (once_tdTP (failTP `adhocTP` decl))                              
   where 
    decl ((Dec (HsFunBind loc matches@((HsMatch _ fun pats rhs ds):ms)))::HsDeclP)
      | pNTtoPN fun == pn
      = do matches' <- mapM addInMatch matches
           if length matches == length matches'   --make sure all adding into matches sucess.
             then return (Dec (HsFunBind loc matches'))
             else mzero   
         
    --add a parameter to a simple pattern binding changing it into a function binding.    
    decl pat@(Dec (HsPatBind loc p rhs ds))
     | patToPName p == pn    
     = do (f,d) <- freeAndDeclaredNames pat
          if elem newParam (f `union` d)
           then error "The new parameter name will cause name clash or semantics change, please select another name!"
           else do addParamInToks p newParam  True
                   newParamPat<-idToPat newParam      
                   addDefaultActualArg True pn (patToPName newParamPat) 
                           (Dec (HsFunBind loc [HsMatch loc (patToPNT p) [newParamPat] rhs ds]))
    decl (Dec (HsPatBind _ p _ _))
     | patToPName p /= pn && elem pn (pNamesInPat p)
      = error "Parameter can not be added to complex pattern binding" 

    decl _=mzero
       
    addInMatch (match@(HsMatch loc  fun  pats rhs  decls)::HsMatchP)
     = do (f,d) <- freeAndDeclaredNames match
          if elem newParam (f `union` d)
           then error "The new parameter name will cause name clash or semantics change, please choose another name!"
           else do addParamInToks fun newParam True
                   newParamPat<-idToPat newParam
                   addDefaultActualArg True pn (patToPName newParamPat) 
                              (HsMatch loc fun (newParamPat:pats) rhs  decls)

addParamInToks t newParam formalParam
  = do ((toks,_),others)<-get   
       let (startPos,endPos) = getStartEndLoc toks t
           (_,toks2) = break (\t->tokenPos t == startPos) toks
           tToks     = takeWhile (\t->tokenPos t<=endPos) toks2
           newToken  = if formalParam  
                         then mkToken Varid startPos $ concatMap tokenCon tToks++" "++newParam  
                         else mkToken Varid startPos $ "("++concatMap tokenCon tToks++" "++newParam ++")"
           toks' = replaceToks toks startPos endPos [newToken]
       put ((toks',modified),others)
                  

--add the default argument declaration right after the declaration defining pn
addDefaultActualArgDecl defaultParamPName parent pn isExported
   =let ds = declsIn parent
    in if not (findEntity pn parent) && not isExported
        then return ds               
        else do argDecl <- defaultArgDecl defaultParamPName  
                let (before,after) = break (defines pn) ds
                     --after can not be empty, no problem with 'head'
                    ds' = before++[head after]++argDecl ++tail after   
                   --Start token stream updating
                ((toks,_),others)<-get   
                let (startPos,endPos) = startEndLocIncFowComment toks (head after) 
                     --divide token stream into two parts.
                    (toks1,toks2) = break (\t->tokenPos t>endPos) toks
                    (_,toks12) = break (\t->tokenPos t>=(fst startPos,0)) toks1 
                    --get the toks defining pn
                    defToks = dropWhile (\t->tokenPos t<startPos) toks12
                    offset = length (concatMap tokenCon (takeWhile (\t->tokenPos t<startPos) toks12))
                    argDecl = pNtoName defaultParamPName++"=undefined"
                    defaultArgToken = mkToken Varid (fst endPos+1,snd startPos) $ whiteSpaces offset++argDecl
                    toks' = if toks1/=[] && endsWithNewLn  (glast "addDefaultActualArgDecl" toks1)
                             then --NO NEED TO ADJUST LAYOUT
                                  toks1++[newLnToken,defaultArgToken,newLnToken]++ compressPreNewLns toks2  
                             else --May need to adjust layout
                                  replaceToks toks startPos endPos (defToks++[newLnToken,newLnToken,defaultArgToken]) 
                put ((toks',modified),others)    
                return ds'         
                
    where
      defaultArgDecl argPName 
       = do let loc1 = pNameToOrig argPName              
            (fileName, int)<-nextValue  --ATTENTION:The first loc1 is not accurate here. should be changed it necessary.
            return [Dec (HsPatBind loc1 (Pat (HsPId (HsVar (PNT argPName  Value (N (Just loc1))))))
                    (HsBody (Exp (HsId (HsVar (PNT (PN (UnQual "undefined") (G (PlainModule "Prelude") "undefined" 
                    (N (Just loc0)))) Value (N (Just (SrcLoc fileName 0 int int )))))))) [])]
 
--suppose function name is f, parameter name is p, then the default argument name is like f_p.    
mkLocalDefaultArgName fun paramName modName t
 =do (f,d)<-freeAndDeclaredNames t  
     vs<-hsVisibleNames fun t
     (fileName,int)<-nextValue
     name<-mkNewName ((pNtoName fun)++"_"++paramName) (nub (f `union` d `union` vs)) 0 
     return (PN (UnQual name) (S (SrcLoc fileName 0 int int)))
       
mkTopLevelDefaultArgName fun paramName fileName modName inscopeNames t
 =do (f,d)<-freeAndDeclaredNames t  
     name<- mkNewName ((pNtoName fun)++"_"++paramName) (nub (f `union` d `union` inscopeNames)) 0 
     let loc =  SrcLoc fileName 0 initValue initValue
     return (PN (UnQual name) (G modName name (N (Just loc))))
   
{-
--add default actual argument to pn in all the calling places.            
addDefaultActualArg pn argPName=applyTP (stop_tdTP (failTP `adhocTP` funApp))
        where 
          funApp (exp@(Exp (HsId (HsVar (PNT pname _ _))))::HsExpP)
              |pname==pn= do --token stream updating
                             addParamInToks exp (pNtoName argPName) False
                             --AST updating
                             argExp<-pNameToExp argPName
                             return (Exp (HsApp exp argExp)) 
          funApp _=mzero               -}
   
addDefaultActualArg recursion pn argPName
        = if recursion then applyTP (stop_tdTP (failTP `adhocTP` funApp))
                       else applyTP (stop_tdTP (failTP `adhocTP` inDecl
                                                       `adhocTP` funApp))
       where 
         inDecl (fun@(Dec (HsFunBind _  ((HsMatch _ (PNT pname _ _) _ _ _):ms)))::HsDeclP)
           | pn == pname 
           = return fun

         inDecl (pat@(Dec (HsPatBind loc1 ps rhs ds))::HsDeclP)
           | pn == patToPName  ps
           = return pat             
         inDecl _ = mzero
     
         funApp (exp@(Exp (HsId (HsVar (PNT pname _ _))))::HsExpP)
          |pname==pn= do --token stream updating
                         addParamInToks exp (pNtoName argPName) False
                         --AST updating
                         argExp<-pNameToExp argPName
                         return (Exp (HsApp exp argExp))          
         funApp _ = mzero

--Add type arg to type siginature 
addArgToSig pn decls
   = let (before,after)=break (\d ->definesType pn d) decls
     in if  after==[] 
       then  return decls
       else  do newSig<-addArgToSig' (head after)  --no problem with head.   
                return (before++newSig++(tail after))

    where
       addArgToSig' sig@(Dec (HsTypeSig loc is c tp))
          =do let tVar=mkNewTypeVarName sig
              typeVar<-newTypeVar tVar tp
              ((toks,_),others)<-get   
              let newSig=if length is==1
                           then  --the type sig only defines the type for pn
                                [Dec (HsTypeSig loc is c typeVar)]  
                           else  --otherwise, seperate it into two type signatures.
                               [Dec (HsTypeSig loc (filter (\x->pNTtoPN x/=pn) is) c tp),  
                                Dec (HsTypeSig loc (filter (\x->pNTtoPN x==pn) is) c typeVar)]
                  (startPos,endPos)=getStartEndLoc toks sig
                  toks1=takeWhile (\t->tokenPos t<startPos) toks
                  (toks11,toks12)=break (\t->tokenPos t>=(fst startPos,0)) toks1
                  offset=length (concatMap tokenCon (takeWhile (\t->tokenPos t<startPos) toks12))
                  t=mkTokensNoFstLnIndent Whitespace  startPos  (astToSource newSig)
                  toks'=replaceToks toks startPos endPos t  
              put ((toks',modified),others)
              return newSig 
     
       --compose a type application using type expressions tv and tp 
       newTypeVar tVar tp
            =do (fileName, int)<-nextValue
                let loc=SrcLoc fileName 0 int int 
                return (Typ (HsTyFun (Typ (HsTyVar (PNT (PN (UnQual tVar) (S loc))
                         (Type (TypeInfo {defType=Nothing, constructors=[], fields=[]})) (N (Just loc))))) tp))

       {- make a fresh type variable name. the new name is the first letter in the alphabet which is not
          used in the type signature. -}
       mkNewTypeVarName sig 
          =mkNewName "a" $ map pNtoName $ (snd.hsTypeVbls undefined) sig
             where mkNewName initName v
                     =if elem initName v
                         then mkNewName ((intToDigit (digitToInt(head initName)+1)):tail initName) v
                         else initName               

addArgInClientMod pnt defaultArg  serverModName ((inscps, exps, mod,ts), fileName)
 = let qual = getImpQualifier pnt inscps
       pn = pNTtoPN pnt
   in if isNothing qual
          then return ((fileName,unmodified), (ts,mod))
          else do (_,((ts',m),(f,_)))<-
                      runStateT (do mod'<-addVarItemInImport serverModName mod pn defaultArg
                                    mod''<-addVarItemInExport mod pn defaultArg
                                    if isInScopeAndUnqualified (pNtoName pn) inscps
                                       then addDefaultActualArgInClientMod  pn  (fromJust qual) defaultArg False mod''
                                       else addDefaultActualArgInClientMod  pn  (fromJust qual) defaultArg True  mod'')
                          ((ts,unmodified), (fileName, initValue))
                  return ((f,m),(ts',mod))  

            
--add default actual argument to pn in all the calling places.            
addDefaultActualArgInClientMod pn qual argPName toBeQualified t
   = applyTP (stop_tdTP (failTP `adhocTP` funApp)) t
  where 
    funApp (exp@(Exp (HsId (HsVar pnt@(PNT pname _ _))))::HsExpP)
      | pname == pn 
       = do vs <- hsVisibleNames pnt t 
            argExp<-if toBeQualified || elem (pNtoName argPName) vs
              then pNameToExp (unqualPNtoQualPN qual argPName)
              else pNameToExp argPName
            addParamInToks exp (pNtoName argPName) False
            return (Exp (HsApp exp argExp)) 
    funApp _=mzero              

myfst (a,_,_,_) = a       
-------------------------------End of adding a parameter-----------------------------------           

-----------------------------------------------------------------------------------------------------
{-Refactoring: Remove a parameter                                                                   
  Description: The refactoring removes a user specified formal parameter in a function binding,and    
               the corresponding actual parameters in all calling places of this function. The
               condition acompanying this refactoring is that the parameter to be removed is not being used.

  --To select a parameter, just stop the cursor at any position between the start and end position of this parameter.
-}
----------------------------------------------------------------------------------------------------- 

rmOneParameter args
 =do let fileName=args!!0
         --fileName'=moduleName fileName
         --modName=Module fileName'   --Assume file name is the same as module name       
         row=read (args!!1)::Int    --but pay attention that people also use unamed module 
         col=read (args!!2)::Int
     modName <-RefacUtils.moduleName fileName
     (inscps, exps, mod, tokList)<-parseSourceFile fileName  
     --pn is the function names. 
     --nth is the nth paramter of pn is to be removed,index starts form 0.
     let (pnt,nth)=getParam tokList (row,col) mod  
         pn=pNTtoPN pnt 
     if pn/=defaultPN 
       then do (mod',(tokList',m))<-doRmParam pn nth mod tokList
               if isExported pnt exps
                then do clients<-clientModsAndFiles modName
                        refactoredClients<-mapM (rmParamInClientMod pnt nth) clients
                        writeRefactoredFiles $ ((fileName,m),(tokList',mod')):refactoredClients                                           
                else  writeRefactoredFiles [((fileName,m), (tokList',mod'))]
       else error "Invalid cursor position!"

--pn: function name; nth: the index of the parameter to be removed. 
doRmParam pn nth mod tokList
       =runStateT (applyTP ((once_tdTP (failTP `adhocTP` inMod
                                              `adhocTP` inMatch
                                              `adhocTP` inPat
                                              `adhocTP` inLet
                                              `adhocTP` inAlt
                                              `adhocTP` inLetStmt)) `choiceTP` failure) mod) (tokList,unmodified)
      where
             --1. pn is declared in top level. 
             inMod (mod@(HsModule loc name exps imps ds):: HsModuleP)  
                |findDecl pn ds False = doRemoving mod  ds
             inMod _ =mzero 

             --2. pn is declared locally in the where clause of a match.
             inMatch (match@(HsMatch loc1 name pats rhs ds)::HsMatchP)
                 | findDecl pn ds False = doRemoving match  ds                                                 
             inMatch _ =mzero

             --3. pn is declared locally in the where clause of a pattern binding.
             inPat (pat@(Dec (HsPatBind loc p rhs ds))::HsDeclP)
                | findDecl pn ds False = doRemoving pat  ds                    
             inPat _=mzero
  
             --4: pn is declared locally in a  Let expression
             inLet (letExp@(Exp (HsLet ds e))::HsExpP)
               | findDecl pn  ds False = doRemoving letExp  ds
             inLet _=mzero
                
             --5. pn is declared locally in a  case alternative.
             inAlt (alt@(HsAlt loc p rhs ds)::HsAltP)
                |findDecl pn ds False = doRemoving  alt  ds
             inAlt _=mzero
      
             --6.pn is declared locally in a let statement.
             inLetStmt (letStmt@(HsLetStmt ds stmts):: HsStmtP)
                |findDecl pn ds False = doRemoving letStmt ds
             inLetStmt _=mzero

             failure=idTP `adhocTP` mod
                     where mod (m::HsModuleP)=error "Refactoring failed"

             doRemoving parent  ds  --PROBLEM: How about doRemoving fails?
                =do rmFormalArg pn nth False True =<<rmNthArgInFunCall pn nth False ds 
                    ds'<-rmNthArgInSig pn nth =<<rmFormalArg pn nth True False  ds  
                    rmNthArgInFunCall pn nth True (replaceDecls parent ds') 

             -- just remove the nth formal parameter.
             rmFormalArg pn nth updateToks checking         
                =applyTP (stop_tdTP (failTP `adhocTP` rmInMatch))
                                                   
               where 
                 rmInMatch (match@(HsMatch loc  fun  pats rhs decls)::HsMatchP) --a formal parameter only exists in a match
                    |pNTtoPN fun==pn  
                   =let  pat=pats!!nth   --get the nth formal parameter
                         pNames=pNamesInPat pat  --get all the names is this pat. (the pat may be just be a variable)
                    in if checking && not ( all (==False) ((map (flip findPName rhs) pNames)) && --not used in rhs
                                            all (==False) ((map (flip findPName decls) pNames))) --not used in the where clause
                         then error  "This parameter can not be removed, as it is used!" 
                         else do if updateToks 
                                   then do (toks,_)<-get
                                           let (startPos,endPos)=getStartEndLoc toks pat
                                               toks'=deleteToks toks startPos endPos
                                           put (toks',modified)
                                           return (HsMatch loc fun (pats\\[pat]) rhs decls)                                  
                                   else return (HsMatch loc fun (pats\\[pat]) rhs decls)                          
                 rmInMatch _=mzero



--remove the nth argument of function pn in all the calling places. the index for the first argument is zero.
rmNthArgInFunCall pn nth updateToks=applyTP (stop_tdTP (failTP `adhocTP` funApp))
        where             
         funApp (exp@(Exp (HsParen (Exp (HsApp e1 e2))))::HsExpP)
              |nth==0 && pn==expToPName e1
             =do if updateToks 
                   then do (toks,_)<-get    --handle the case like '(fun x) => fun "
                           let (before,after)=splitAt (nth+1) (unfoldHsApp exp) --remove the nth argument
                               (startPos,endPos)=getStartEndLoc toks exp
                               e1Toks=getToks (getStartEndLoc toks e1) toks
                               toks'=replaceToks toks startPos endPos e1Toks
                           put (toks',modified)
                           return (e1)     
                   else return (e1)
         funApp (exp@(Exp (HsApp e1 e2))::HsExpP)
              |pn==(expToPName.(ghead "rmNthArgInFunCall").unfoldHsApp) exp  --test if this application is a calling of fun pn.
               =do let (before,after)=splitAt (nth+1) (unfoldHsApp exp) --remove the nth argument
                   if updateToks 
                    then do (toks,_)<-get
                            let (startPos,endPos)=getStartEndLoc toks (ghead "rmNthArgInFunCall" after)  
                                toks'=deleteToks toks startPos endPos
                            put (toks',modified) 
                            return (foldHsApp (before++tail after))  
                    else return (foldHsApp (before++tail after))  --reconstruct the function application.        
                 
         funApp _=mzero                  
        
         --deconstruct a function application into a list of expressions.
         unfoldHsApp exp=
              case exp of
                  (Exp (HsApp e1 e2))->(unfoldHsApp e1)++[e2] 
                  _ ->[exp]
         --reconstruct  a function application by a list of expressions.                           
         foldHsApp exps=foldl1 (\e1 e2->(Exp (HsApp e1 e2))) exps 


rmNthArgInSig pn nth decls
 = let (before,after)=break (\d ->definesType pn d) decls
   in if  after==[] 
       then  return decls
       else  do newSig<-rmNthArgInSig' nth  (head after)  --no problem with 'head'
                return (before++newSig++(tail after))

   where rmNthArgInSig' nth sig@(Dec (HsTypeSig loc is c tp))
          =do (toks,_)<-get
              let newSig=if length is ==1  
                            then --this type signature only defines the type of pn 
                                 [Dec (HsTypeSig loc is c (rmNth tp))]
                            else --this type signature also defines the type of other ids.
                                 [Dec (HsTypeSig loc (filter (\x->pNTtoPN x/=pn) is) c tp),  
                                  Dec (HsTypeSig loc (filter (\x->pNTtoPN x==pn) is) c (rmNth tp))]
                  (startPos,endPos)=getStartEndLoc toks sig 
                  t=mkTokensNoFstLnIndent Whitespace startPos (astToSource newSig)
                  toks'=replaceToks toks startPos endPos t   
              put (toks',modified)
              return newSig

         rmNth tp=let (before,after)=splitAt nth (unfoldHsTypApp tp)
                    in  (foldHsTypApp (before ++ tail after))
            
         --deconstruct a type application into a list of types
         unfoldHsTypApp typ =
               case typ of (Typ (HsTyFun t1 t2)) ->t1:unfoldHsTypApp t2
                           _ ->[typ]

         --reconstruct a type application by a list of type expression.
         foldHsTypApp ts=foldr1 (\t1 t2->(Typ (HsTyFun t1 t2))) ts
 
--get the function name and the index of the parameter to be removed from the cursor position.
getParam toks pos  
     =(fromMaybe (defaultPNT, 0)).(applyTU (once_tdTU (failTU `adhocTU` inMatch)))
    where
       inMatch (match@(HsMatch loc  fun  pats rhs  decls)::HsMatchP)
          |inRange pos (getStartEndLoc toks pats) 
          =let paramPosRanges=map (getStartEndLoc toks) pats 
               elem=find (inRange pos) paramPosRanges
               in if isNothing elem 
                    then error "Invalid cursor position!"  -- cursor doesn't stop at a parameter position.
                     else Just (fun, fromJust (elemIndex (fromJust elem) paramPosRanges))
       inMatch _=Nothing
      
       inRange pos (startPos,endPos)=pos>=startPos && pos<=endPos
        
rmParamInClientMod pnt nth (m, fileName)       
 = do (inscps, exps, mod, ts) <-parseSourceFile fileName
      let qualifier = getImpQualifier pnt inscps
          pn = pNTtoPN pnt
      if isNothing qualifier
          then return ((fileName,unmodified), (ts,mod))
          else do (mod',(ts',m)) <- runStateT (rmNthArgInFunCall pn nth True mod) (ts,unmodified)
                  return ((fileName,m), (ts',mod')) 

