

module RefacRenaming(rename) where

import Prelude hiding (putStrLn)
import AbstractIO (putStrLn)
import PosSyntax
import Maybe
import TypedIds
import UniqueNames 
import List  
import PNT 
import RefacUtils 

{-This refactoring renames an indentifier to a user-specified name.
  Conditions: a: the indentifier to be renamed should be defined in the current module.
              b. the user provided name should be a valid name with regard to the name space of the identifier.
              c. the new name should not change the semantics of the program, and should not cause any name 
                 clash/conflict/ambiguity problem in the program. 
  Attention: a. To select an identifier, stop the cursor at the beginning position of any occurrence of the identifier.
             b. Renaming a qualified name will not change the qualifier; 
             c. In current module, an unqualified name won't become qualified after renaming; but, in client modules,
                an unqualified name might become qualified after renaming to avoid ambiguity prolem. In case the new name,
                say 'f', will cause ambiguous occurrence in the current module (this is because the identifier 'f' is
                imported from other modules), the user will be prompted to choose another new name or qualify the use 
                of 'f' before doing renaming.                      
 -}

{-In the current implementation, we assume that module name is same as the file name, but we
  should keep in mind that people also use unnamed modules.
-}
rename args
 = do let fileName = args!!0
          --fileName' = moduleName fileName
          --modName   = Module fileName'  
          newName   = args!!1            
          row       = read (args!!2)::Int
          col       = read (args!!3)::Int
      modName <- RefacUtils.moduleName fileName
      (inscps, exps, mod,tokList) <- parseSourceFile fileName    
      let pnt       = locToPNT fileName row col mod   
          defineMod = if isTopLevelPN (pNTtoPN pnt)
                           then (fromJust (hasModName pnt))
                           else modName
      unless (pnt /= defaultPNT  && isValidNewName pnt newName) 
         $ error "Invalid cursor position or the new name is invalid!\n" 
      unless (pNTtoName pnt/=newName) $ error "The new name is same as the old name" 
      unless (defineMod == modName ) ( error ("This idenifier is defined in module " ++ showMod defineMod ++ 
                                             ", please do renaming in that module!"))
      if isMainModule modName && pNTtoName pnt == "main"
          then error ("The 'main' function defined in a 'Main' module should not be renamed!")
          else do (mod',((tokList',modified),_))<-doRenaming pnt newName modName mod inscps exps
                                                  ((tokList,unmodified),(fileName,initValue))
                  if isExported pnt exps  --no matter whether this pn is used or not.
                     then do clients<-clientModsAndFiles modName  --returns [(module name ,filename)] 
                             refactoredClients <- mapM (renameInClientMod pnt newName) clients                           
                             writeRefactoredFiles $ ((fileName,modified),(tokList',mod')):refactoredClients
                     else writeRefactoredFiles [((fileName,modified), (tokList',mod'))]

---------Rename a type constructor-------------------------- 
doRenaming oldPNT@(PNT (PN _ (G _ _ _)) (Type _) _) newName modName mod inscps exps env
  =doRenaming' oldPNT newName  mod  (hsTypeConstrsandClasses modName) exps env

---------Rename a class name---------------------------------
doRenaming oldPNT@(PNT _ (Class _) _) newName modName mod inscps exps env 
  =doRenaming' oldPNT newName  mod (hsTypeConstrsandClasses modName) exps env 

---------Rename a data structure-----------------------------
doRenaming oldPNT@(PNT _ (ConstrOf _  _) _) newName modName mod inscps exps env
  =doRenaming' oldPNT newName mod  (hsDataConstrs modName)  exps env

---------Rename an instance name-----------------------------
doRenaming oldPNT@(PNT _ (MethodOf i _) _) newName modName mod inscps exps env -- Claus
   =runStateT (renameTopLevelVarName oldPNT newName modName inscps  exps mod True True) env

---------Rename a type variable name------------------------------
doRenaming oldPNT@(PNT oldPN@(PN i (S loc)) (Type _) _) newName modName mod inscps exps env
   = do let (before, parent, after)=divideDecls (declsIn mod) oldPN
        (parent',toks)<-doRenaming' oldPNT newName parent (hsTypeVbls modName) exps  env
        return ((replaceDecls mod (before++parent'++after)),toks) 

---------Rename a field name -------------------------------------
doRenaming oldPNT@(PNT _ (FieldOf con typeInfo) _) newName modName mod inscps exps env 
  =do let (before,parent,after)=divideDeclsAccordingToPNT (declsIn mod) oldPNT
          siblingNames=siblingFieldNames oldPNT
      (f',d') <- freeAndDeclaredVbls (before++after)  
      if elem newName (siblingNames ++ map pNtoName d')
        then error("Name '"++newName++"'  already existed\n")
        else if elem newName $ fieldNames oldPNT
               then runStateT (renameTopLevelVarName oldPNT newName modName inscps exps mod False False) env
               else runStateT (renameTopLevelVarName oldPNT newName modName inscps exps mod False True) env
--------Rename a value varaible name--------------------------------
doRenaming oldPNT@(PNT oldPN Value loc) newName modName mod inscps exps env 
 = runStateT (applyTP (once_buTP (failTP `adhocTP` renameInMod
                             `adhocTP` renameInMatch
                             `adhocTP` renameInPattern
                             `adhocTP` renameInExp
                             `adhocTP` renameInAlt
                             `adhocTP` renameInStmts)) mod) env
   where 
     -- 1. The name is declared in a module(top level name)
     renameInMod (mod::HsModuleP)
        |isDeclaredIn oldPN mod=renameTopLevelVarName oldPNT newName modName inscps exps mod True True
     renameInMod mod=mzero

     -- 2. The name is declared in a match
     renameInMatch (match::HsMatchP)
        |isDeclaredIn oldPN match=renameLocalVarName oldPN newName match
     renameInMatch match=mzero

     -- 3. The name is declared in a pattern binding.
     renameInPattern (pat:: HsDeclP)
        |isDeclaredIn oldPN pat=renameLocalVarName oldPN newName pat
     renameInPattern pat=mzero
 
     --4.The name is declared in a expression
     renameInExp (exp::HsExpP)
       |isDeclaredIn oldPN exp=renameLocalVarName oldPN newName exp
     renameInExp exp=mzero

     --5.The name is declared in a case alternative
     renameInAlt (alt::HsAltP)
       |isDeclaredIn oldPN alt=renameLocalVarName oldPN newName alt
     renameInAlt alt=mzero

     --6: The name is declared in a statment expression
     renameInStmts (stmts::HsExpP)
       |isDeclaredIn oldPN stmts=renameLocalVarName oldPN newName stmts
     renameInStmts stmts=mzero


doRenaming' oldPNT@(PNT oldPN _ _) newName ast fun exps env
 = let (f',d')= fun ast
       (f,d)  =((nub.map pNtoName.filter (not.isQualifiedPN)) f', (nub.map pNtoName) d')
   in if elem newName (d \\ [pNtoName oldPN])
         then error("Name '"++newName++"'  already existed\n")
         else if elem newName f 
               then error ("The new name will cause ambiguous occurrence problem,"
                           ++" please select another new name or qualify the use of "
                           ++ newName ++ " before renaming!\n")             
               else if causeNameClashInExports oldPNT  newName exps
                    then error ("The new name will cause cause conflicting exports, please select another new name!")
                    else runStateT (renamePNwithUpdLexer oldPN newName ast) env 
       
renameTopLevelVarName oldPNT@(PNT oldPN _ _) newName modName inscps exps mod existChecking exportChecking
  =do -- f' contains names imported from other modules; 
      -- d' contains the top level names declared in this module;  
     (f',d') <- freeAndDeclaredVbls mod  
      --filter those qualified free variables in f'
     let (f,d) = ((nub.map pNtoName.filter (not.isQualifiedPN)) f', (nub.map pNtoName) d')
     if elem newName f
       then error ("The new name will cause ambiguous occurrence problem,"
                   ++" please select another new name or qualify the use of ' " 
                   ++ newName ++ "' before renaming!\n")  -- Another implementation option is to add the qualifier
                                                         -- to newName automatically.
       else if existChecking && elem newName (d \\ [pNtoName oldPN])  --only check the declared names here.
             then error ("Name '"++newName++"'  already existed\n") --the same name has been declared in this module.
             else if exportChecking && causeNameClashInExports oldPNT newName exps
                    then error ("The new name will cause cause conflicting exports, please select another new name!") 
                    else if exportChecking && causeAmbiguityInExports oldPN  newName inscps mod
                          then error $"The new name will cause ambiguity in the exports of module "++ show modName ++ ", please select another name!"   
                          else do  -- get all of those declared names visible to oldPN at where oldPN is used.
                                 ds<-hsVisibleNames oldPN mod
                                 -- '\\[pNtoName oldPN]' handles the case in which the new name is same as the old name   
                                 if existChecking && elem newName ((nub (ds `union` f)) \\[pNtoName oldPN])
                                   then error ("Name '"++newName++"'  already existed, or rename '"
                                                ++pNtoName oldPN++ "' to '"++newName++
                                                "' will change the program's semantics!\n")
                                   else if exportChecking && isInScopeAndUnqualified newName inscps
                                          then renamePNwithUpdLexer1 oldPN modName newName mod
                                          else renamePNwithUpdLexer oldPN newName mod 

renameLocalVarName oldPN newName t  
  =do (f,d) <- freeAndDeclaredNames t
      if elem newName (d \\ [pNtoName oldPN])  --only check the declared names here.
        then error ("Name '"++newName++"'  already existed\n")
        else do -- get all of those declared names visible to oldPN at where oldPN is used.
                ds<-hsVisibleNames oldPN t
                -- '\\[pNtoName oldPN]' handles the case in which the new name is same as the old name   
                if elem newName ((nub (ds `union` f)) \\[pNtoName oldPN])
                  then error ("Name '"++newName++"'  already existed, or rename '"
                               ++pNtoName oldPN++ "' to '"++newName++
                              "' will change the program's semantics!\n")
                  else renamePNwithUpdLexer oldPN newName t 
      
 
   
renameInClientMod pnt@(PNT oldPN _ _) newName (m, fileName)
  =do (inscps, exps ,mod ,ts) <-parseSourceFile fileName
      let qualifier=getImpQualifier pnt inscps 
      if isNothing qualifier --this name is not imported, but it maybe used in the import list
       then 
        do (mod', ((ts',m),(f,_)))<-runStateT (renamePNwithUpdLexer oldPN newName mod) ((ts,unmodified), (fileName, initValue))   
           return ((f,m),(ts',mod'))  
       else
         do if causeNameClashInExports pnt newName exps  
             then error $"The new name will cause conflicting exports in module "++ show m ++ ", please select another name!"
             else do (mod',((ts',modified),(f,_)))<-runStateT (do mod'<-qualifyTopLevelVar newName inscps mod
                                                                  worker oldPN newName mod' inscps (fromJust qualifier))
                                                           ((ts,unmodified),(fileName,initValue))
                     return ((f,modified), (ts',mod')) 
          
  where
     qualifyTopLevelVar name inscps t
       = applyTP(full_tdTP(adhocTP idTP atPnt)) t  
        where
           atPnt pnt@(PNT pn ty loc@(N (Just (SrcLoc fileName _  row col))))
              | isTopLevelPNT pnt &&  pNTtoName pnt == name && pNTinUseLoc pnt 
                 && not (findPntInLhs pnt t || findPntInImp pnt t)
               = do ((toks,_),others)<-get   
                    let qualifier=fromJust $ getImpQualifier pnt inscps 
                        toks'=replaceToks toks (row,col) (row,col) 
                          [mkToken Varid  (row,col) (pNameToStr (replaceNameInPN1 qualifier pn name))]
                    put ((toks',modified),others)
                    return (PNT (replaceNameInPN1 qualifier pn name) ty loc)
           atPnt x = return x

     worker oldPN newName mod inscps qual  
        = do  vs<-hsVisibleNames oldPN mod   --Does this check names other than variable names?
              if elem newName ((nub vs) \\[pNtoName oldPN])  || isInScopeAndUnqualified newName inscps
                 then renamePNwithUpdLexer1 oldPN qual newName mod  --rename to qualified Name
                 else do renamePNwithUpdLexer oldPN newName mod 

isValidNewName oldName newName
 = case oldName of
        (PNT (PN i (G _ _ _)) (Type _) _) ->if isConId newName then True
                                                               else error "Invalid type constructor name!"  
        (PNT _ (Class _) _)               ->if isConId newName then True
                                                               else error "Invalid class name!"
        (PNT _ (ConstrOf _  _) _)         ->if isConId newName then True
                                                               else error "Invalid data constructor name!"

        (PNT _ (FieldOf _ _ ) _ )         ->if isVarId newName then True
                                                               else error "Invalid field name!" 

        (PNT _ (MethodOf i _) _)          ->if isVarId newName then True 
                                                               else error "Invalid class instance name!"
        
        (PNT (PN i (S loc)) (Type _) _)   ->if isVarId newName then True
                                                               else error "Invalid type variable name!"
 
        (PNT oldPN Value loc)             ->let oldName' = pNTtoName oldName
                                            in if isVarId oldName' && not (isVarId newName)
                                                 then error "The new name should be an identifier!"
                                                 else if isVarSym oldName' && not (isVarSym newName)
                                                     then error "The new name should be an operator!"
                                                     else if (isVarId oldName' && isVarId newName) || 
                                                             (isVarSym oldName' && isVarSym newName)
                                                           then True
                                                           else error "Invalid new name!"

--test whether the new name will cause name clash in the exports--
--ExpList :: [(n,Ent n)]-- eg: [("f",Ent "A" (HsVar "f") Value)]

causeNameClashInExports oldName@(PNT pn t1 _) newName exps
  = case hasModName pn of
         Just modName -> if isExported oldName exps 
                          then isJust $ find (\x->show(fst x)==show newName) exps   
                          else False
         Nothing     -> False  --oldName is a local name 

showMod (PlainModule name) = name

siblingFieldNames (PNT pn (FieldOf con typeInfo) loc)
  = let cons=constructors typeInfo
        orig1 = pNtoLoc pn
        con=filter (\x->elem orig1 (map pNtoLoc (fromMaybe [] (conFields x)))) cons
    in if con==[] then []
                  else (map pNtoName1 (fromMaybe [] (conFields (head con))))\\ [pNtoName pn]
   where
     pNtoLoc (PN _ orig1) = orig1
     pNtoName1 (PN i _) = i

fieldNames  (PNT pn (FieldOf con typeInfo) _)
 =let cons=constructors typeInfo
  in  (nub. map pNtoName) $ concatMap (fromMaybe [].conFields) cons 
  where  pNtoName (PN i _) = i




